module Main exposing (main)

import Html exposing (..)
import Html.Attributes exposing (srcdoc)
import Playground as P exposing (..)
import Tuple exposing (first, second)


type alias Memory =
    { dx : Number
    , count : Int
    }


type alias Bounds =
    { ymin : Number
    , ymax : Number
    , xmin : Number
    , xmax : Number
    }


type alias Coords =
    ( Number, Number )


init : Memory
init =
    { dx = 0.8
    , count = 0
    }


main = game


game =
    P.game view update init


view : Computer -> Memory -> List Shape
view computer memory =
    let
        bounds =
            { ymin = 0
            , ymax = 32 * computer.screen.height / computer.screen.width
            , xmin = 0
            , xmax = 32
            }

        at =
            placeAtCoords computer bounds

        atCorner =
            placeCornerAtCoords computer bounds

        fx x = x * x * x / 32 / 32
        corner = case memory.count // 30 of
            0 ->
                TopLeft
            1 -> TopRight
            _ -> TopLeft
    in
    List.append
        (makeRectangles computer atCorner fx bounds memory [])
        <|
            List.append
                (makeCircles computer fx bounds memory [])
                ([ P.words black "Welcome to my math project! Left and right arrows adjust dx, up and down change the visualization."
                    |> at ( (bounds.xmax - bounds.xmin) / 2, bounds.ymax - 0.5)
                , P.words black ("The current dx is " ++ (String.fromFloat memory.dx))
                    |> at ( (bounds.xmax - bounds.xmin) / 2, bounds.ymax - 1)
                , P.words black ("The left approximation is " ++ (String.fromFloat <| (\n->n/32/32) <| getArea computer TopLeft fx bounds memory))
                    |> at ( (bounds.xmax - bounds.xmin) / 2, bounds.ymax - 1.5)
                , P.words black ("The right approximation is about " ++ (String.fromFloat <| (\n->n/32/32) <| getArea computer TopRight fx bounds memory))
                    |> at ( (bounds.xmax - bounds.xmin) / 2, bounds.ymax - 2)
                ])


fillLeftFxs : Computer -> (Number -> Number) -> Bounds -> Memory -> List Number -> List Number
fillLeftFxs computer fx bounds memory fxs =
    let
        xcoord = toFloat (List.length fxs) * memory.dx
        ycoord = fx xcoord
        widthOfCell = computer.screen.width / (bounds.xmax - bounds.xmin)
        newVal = memory.dx * widthOfCell * ycoord * widthOfCell
    in
    if memory.dx + memory.dx * (toFloat <| List.length fxs) < bounds.xmax then
        fillLeftFxs computer fx bounds memory <| List.append fxs <| [newVal]
    else
        List.append fxs <| [(bounds.xmax - memory.dx * (toFloat <| List.length fxs)) * widthOfCell * ycoord * widthOfCell]



fillRightFxs : Computer -> (Number -> Number) -> Bounds -> Memory -> List Number -> List Number
fillRightFxs computer fx bounds memory fxs =
    let
        xcoord = toFloat (List.length fxs) * memory.dx
        ycoord = fx <| memory.dx + xcoord
        widthOfCell = computer.screen.width / (bounds.xmax - bounds.xmin)
        newVal = memory.dx * widthOfCell * ycoord * widthOfCell
    in
    if memory.dx + memory.dx * (toFloat <| List.length fxs) < bounds.xmax then
        fillRightFxs computer fx bounds memory <| List.append fxs <| [newVal]
    else
        List.append fxs <| [(bounds.xmax - memory.dx * (toFloat <| List.length fxs)) * widthOfCell * ycoord * widthOfCell]


getArea : Computer -> Corner -> (Number -> Number) -> Bounds -> Memory -> Number
getArea computer corner fx bounds memory =
    let
        fxs0 = fillLeftFxs computer fx bounds memory []
        fxs1 = fillRightFxs computer fx bounds memory []
    in
    case corner of
        TopLeft ->
            List.sum fxs0
        TopRight ->
            List.sum fxs1
                
        


makeCircles : Computer -> (Number -> Number) -> Bounds -> Memory -> List Shape -> List Shape
makeCircles computer fx bounds memory prev =
    let
        resolution = (bounds.xmax - bounds.xmin) / 512
        xcoord = toFloat (List.length prev) * resolution
        ycoord = fx xcoord
        newCirc =
            circle purple 2
                |> placeAtCoords computer bounds (xcoord, ycoord)
    in
    if
        xcoord < bounds.xmax
    then
        makeCircles computer fx bounds memory <| List.append prev [ newCirc ]
    else
        prev



makeRectangles : Computer -> (Coords -> Corner -> (Number, Number) -> Shape -> Shape) -> (Number -> Number) -> Bounds -> Memory -> List Shape -> List Shape
makeRectangles computer atCorner fx bounds memory prev =
    let
        xcoord = toFloat (List.length prev) * memory.dx
        ycoord = fx xcoord
        widthOfCell = computer.screen.width / (bounds.xmax - bounds.xmin)
        corner = case (memory.count // 30) of
            0 ->
                TopLeft
            1 ->
                TopRight
            _ ->
                TopLeft
        newRec = 
            rectangle red (memory.dx * widthOfCell) (ycoord * widthOfCell)
                |> atCorner (xcoord, ycoord) corner (memory.dx, ycoord)
    in
    if
        xcoord - memory.dx < bounds.xmax
    then
        makeRectangles computer atCorner fx bounds memory <| List.append prev [ newRec ]

    else
        prev


update : Computer -> Memory -> Memory
update computer memory =
    let
        step = 0.02
        threshold = 0.02
    in
    { memory
        | dx = 
            (toFloat
            <| round
            <| toFloat 100 * (
                if memory.dx > threshold then
                    memory.dx + step * toX computer.keyboard
                else
                    computer.keyboard
                        |> toX
                        |> max 0
                        |> (*) step
                        |> (+) threshold
            )) / toFloat 100
        , count =
                if memory.count <= 0 then
                    memory.count + 50
                else
                    if memory.count >= 60 then
                        memory.count - 50
                    else
                        (+) memory.count <| round <| toY computer.keyboard
    }


placeAtScreen : Computer -> Coords -> Shape -> Shape
placeAtScreen computer coords shape =
    let
        x =
            first coords

        y =
            second coords
    in
    shape |> P.move (-computer.screen.width / 2 + x) (-computer.screen.height / 2 + y)


placeAtCoords : Computer -> Bounds -> Coords -> Shape -> Shape
placeAtCoords computer bounds (x, y) shape =
    placeAtScreen
        computer 
        ( computer.screen.width * x / (bounds.xmax - bounds.xmin)
        , computer.screen.height * y / (bounds.ymax - bounds.ymin)
        )
        shape


type Corner = TopLeft | TopRight

placeCornerAtCoords : Computer -> Bounds -> Coords -> Corner -> (Number, Number) -> Shape -> Shape
placeCornerAtCoords computer bounds (x, y) corner (width, height) shape =
    shape
        |> placeAtCoords computer bounds
            (case corner of
                TopLeft ->
                    (x + width / 2, y - height / 2)
                TopRight ->
                    (x - width / 2, y - height / 2)
            )
